@@ -47,11 +47,10 @@
 #include <pwd.h>
 #include <sys/utsname.h>
 #include <sys/types.h>
-#include <openssl/ui.h>
+#include <termios.h>
 #ifdef LIBPROXY_HDR
 #include LIBPROXY_HDR
 #endif
-
 #include <getopt.h>
 
 #include "openconnect-internal.h"
@@ -63,7 +62,10 @@ static void write_progress(void *_vpninfo,
 static void syslog_progress(void *_vpninfo,
 			    int level, const char *fmt, ...);
 static int validate_peer_cert(void *_vpninfo,
-			      X509 *peer_cert, const char *reason);
+			      OPENCONNECT_X509 *peer_cert,
+			      const char *reason);
+static int process_auth_form(void *_vpninfo,
+			     struct oc_auth_form *form);
 
 /* A sanity check that the openconnect executable is running against a
    library of the same version */
@@ -76,9 +78,12 @@ int background;
 int do_passphrase_from_fsid;
 int nocertcheck;
 int non_inter;
+int cookieonly;
 
 enum {
-	OPT_AUTHGROUP = 0x100,
+	OPT_AUTHENTICATE = 0x100,
+	OPT_AUTHGROUP,
+	OPT_BASEMTU,
 	OPT_CAFILE,
 	OPT_CONFIGFILE,
 	OPT_COOKIEONLY,
@@ -102,6 +107,7 @@ enum {
 	OPT_SERVERCERT,
 	OPT_USERAGENT,
 	OPT_NON_INTER,
+	OPT_DTLS_LOCAL_PORT,
 };
 
 #ifdef __sun__
@@ -127,11 +133,11 @@ static struct option long_options[] = {
 	OPTION("help", 0, 'h'),
 	OPTION("interface", 1, 'i'),
 	OPTION("mtu", 1, 'm'),
+	OPTION("base-mtu", 1, OPT_BASEMTU),
 	OPTION("setuid", 1, 'U'),
 	OPTION("script", 1, 's'),
 	OPTION("script-tun", 0, 'S'),
 	OPTION("syslog", 0, 'l'),
-	OPTION("key-type", 1, 'K'),
 	OPTION("key-password", 1, 'p'),
 	OPTION("proxy", 1, 'P'),
 	OPTION("user", 1, 'u'),
@@ -140,6 +146,7 @@ static struct option long_options[] = {
 	OPTION("cafile", 1, OPT_CAFILE),
 	OPTION("config", 1, OPT_CONFIGFILE),
 	OPTION("no-dtls", 0, OPT_NO_DTLS),
+	OPTION("authenticate", 0, OPT_AUTHENTICATE),
 	OPTION("cookieonly", 0, OPT_COOKIEONLY),
 	OPTION("printcookie", 0, OPT_PRINTCOOKIE),
 	OPTION("quiet", 0, 'q'),
@@ -163,6 +170,7 @@ static struct option long_options[] = {
 	OPTION("no-cert-check", 0, OPT_NO_CERT_CHECK),
 	OPTION("force-dpd", 1, OPT_FORCE_DPD),
 	OPTION("non-inter", 0, OPT_NON_INTER),
+	OPTION("dtls-local-port", 1, OPT_DTLS_LOCAL_PORT),
 	OPTION(NULL, 0, 0)
 };
 
@@ -172,17 +180,53 @@ static void helpmessage(void)
 		 "  http://www.infradead.org/openconnect/mail.html\n"));
 }
 
+static void print_build_opts(void)
+{
+	const char *comma = ", ", *sep = comma + 1;
+
+#if defined (OPENCONNECT_OPENSSL)
+	printf(_("Using OpenSSL. Features present:"));
+#elif defined (OPENCONNECT_GNUTLS)
+	printf(_("Using GnuTLS. Features present:"));
+#endif
+
+	if (openconnect_has_tss_blob_support()) {
+		printf("%sTPM", sep);
+		sep = comma;
+	}
+#if defined (OPENCONNECT_OPENSSL) && defined (HAVE_ENGINE)
+	else {
+		printf("%sTPM (%s)", sep, _("OpenSSL ENGINE not present"));
+		sep = comma;
+	}
+#endif
+	if (openconnect_has_pkcs11_support()) {
+		printf("%sPKCS#11", sep);
+		sep = comma;
+	}
+
+#ifdef HAVE_DTLS
+	printf("%sDTLS", sep);
+#if defined (OPENCONNECT_GNUTLS) && defined (DTLS_OPENSSL)
+	printf(" (%s)", _("using OpenSSL"));
+#endif
+	printf("\n");
+#else
+	printf(_("\nWARNING: No DTLS support in this binary. Performance will be impaired.\n"));
+#endif
+}
+
 static void usage(void)
 {
 	printf(_("Usage:  openconnect [options] <server>\n"));
 	printf(_("Open client for Cisco AnyConnect VPN, version %s\n\n"), openconnect_version_str);
+	print_build_opts();
 	printf("      --config=CONFIGFILE         %s\n", _("Read options from config file"));
 	printf("  -b, --background                %s\n", _("Continue in background after startup"));
-	printf("      --pid-file=PIDFILE          %s\n", _("Write the daemons pid to this file"));
+	printf("      --pid-file=PIDFILE          %s\n", _("Write the daemon's PID to this file"));
 	printf("  -c, --certificate=CERT          %s\n", _("Use SSL client certificate CERT"));
 	printf("  -e, --cert-expire-warning=DAYS  %s\n", _("Warn when certificate lifetime < DAYS"));
 	printf("  -k, --sslkey=KEY                %s\n", _("Use SSL private key file KEY"));
-	printf("  -K, --key-type=TYPE             %s\n", _("Private key type (PKCS#12 / TPM / PEM)"));
 	printf("  -C, --cookie=COOKIE             %s\n", _("Use WebVPN cookie COOKIE"));
 	printf("      --cookie-on-stdin           %s\n", _("Read cookie from standard input"));
 	printf("  -d, --deflate                   %s\n", _("Enable compression (default)"));
@@ -196,6 +240,7 @@ static void usage(void)
 	printf("      --csd-user=USER             %s\n", _("Drop privileges during CSD execution"));
 	printf("      --csd-wrapper=SCRIPT        %s\n", _("Run SCRIPT instead of CSD binary"));
 	printf("  -m, --mtu=MTU                   %s\n", _("Request MTU from server"));
+	printf("      --base-mtu=MTU              %s\n", _("Indicate path MTU to/from server"));
 	printf("  -p, --key-password=PASS         %s\n", _("Set key passphrase or TPM SRK PIN"));
 	printf("      --key-password-from-fsid    %s\n", _("Key passphrase is fsid of file system"));
 	printf("  -P, --proxy=URL                 %s\n", _("Set proxy server"));
@@ -214,6 +259,7 @@ static void usage(void)
 	printf("  -v, --verbose                   %s\n", _("More output"));
 	printf("  -x, --xmlconfig=CONFIG          %s\n", _("XML config file"));
 	printf("      --authgroup=GROUP           %s\n", _("Choose authentication login selection"));
+	printf("      --authenticate              %s\n", _("Authenticate only and print login info"));
 	printf("      --cookieonly                %s\n", _("Fetch webvpn cookie only; don't connect"));
 	printf("      --printcookie               %s\n", _("Print webvpn cookie before connecting"));
 	printf("      --cafile=FILE               %s\n", _("Cert file for server verification"));
@@ -228,6 +274,7 @@ static void usage(void)
 	printf("      --reconnect-timeout         %s\n", _("Connection retry timeout in seconds"));
 	printf("      --servercert=FINGERPRINT    %s\n", _("Server's certificate SHA1 fingerprint"));
 	printf("      --useragent=STRING          %s\n", _("HTTP header User-Agent: field"));
+	printf("      --dtls-local-port=PORT      %s\n", _("Set local port for DTLS datagrams"));
 	printf("\n");
 
 	helpmessage();
@@ -263,7 +310,18 @@ static void handle_sigusr(int sig)
 static FILE *config_file = NULL;
 static int config_line_num = 0;
 
-#define keep_config_arg() (config_arg)
+/* There are three ways to handle config_arg:
+ *
+ * 1. We only care about it transiently and it can be lost entirely
+ *    (e.g. vpninfo->reconnect_timeout = atoi(config_arg);
+ * 2. We need to kep it, but it's a static string and will never be freed
+ *    so when it's part of argv[] we can use it in place, but when it comes
+ *    from a file we have to strdup() because otherwise it'll be overwritten.
+ *    For this we use the keep_config_arg() macro below.
+ * 3. It may be freed during normal operation, so we have to use strdup()
+ *    even when it's an option from argv[]. (e.g. vpninfo->cert_password).
+ */
+#define keep_config_arg() (config_file?strdup(config_arg):config_arg)
 
 static int next_option(int argc, char **argv, char **config_arg)
 {
@@ -280,7 +338,7 @@ static int next_option(int argc, char **argv, char **config_arg)
  next:
 	if (!config_file) {
 		opt = getopt_long(argc, argv,
-				  "bC:c:e:Ddg:hi:k:K:lpP:Q:qSs:U:u:Vvx:",
+				  "bC:c:e:Ddg:hi:k:lpP:Q:qSs:U:u:Vvx:",
 				  long_options, NULL);
 
 		*config_arg = optarg;
@@ -359,26 +417,11 @@ static int next_option(int argc, char **argv, char **config_arg)
 
 }
 
-static int null_ui_open(UI *ui)
-{
-	fprintf(stderr, _("Cannot ask for user input while in non-interactive mode\n"));
-	fflush(stderr);
-	return 0;
-}
-
-static void disable_openssl_ui()
-{
-	UI_METHOD *ui_method = UI_create_method((char *)"OpenConnect non-interactive UI");
-	UI_method_set_opener(ui_method, null_ui_open);
-	UI_set_default_method(ui_method);
-}
-
 int main(int argc, char **argv)
 {
 	struct openconnect_info *vpninfo;
 	struct utsname utsbuf;
 	struct sigaction sa;
-	int cookieonly = 0;
 	int use_syslog = 0;
 	char *urlpath = NULL;
 	char *proxy = getenv("https_proxy");
@@ -400,7 +443,7 @@ int main(int argc, char **argv)
 			openconnect_binary_version, openconnect_version_str);
 	}
 			
-	openconnect_init_openssl();
+	openconnect_init_ssl();
 
 	vpninfo = malloc(sizeof(*vpninfo));
 	if (!vpninfo) {
@@ -412,15 +455,23 @@ int main(int argc, char **argv)
 	/* Set up some defaults */
 	vpninfo->tun_fd = vpninfo->ssl_fd = vpninfo->dtls_fd = vpninfo->new_dtls_fd = -1;
 	vpninfo->useragent = openconnect_create_useragent("Open AnyConnect VPN Agent");
-	vpninfo->mtu = 1406;
+	vpninfo->reqmtu = 0;
 	vpninfo->deflate = 1;
 	vpninfo->dtls_attempt_period = 60;
 	vpninfo->max_qlen = 10;
 	vpninfo->reconnect_interval = RECONNECT_INTERVAL_MIN;
 	vpninfo->reconnect_timeout = 300;
+	vpninfo->uid_csd = 0;
+	/* We could let them override this on the command line some day, perhaps */
+#ifdef __APPLE__
+	vpninfo->csd_xmltag = "csdMac";
+#else
+	vpninfo->csd_xmltag = "csdLinux";
+#endif
 	vpninfo->uid_csd = 0;
 	vpninfo->uid_csd_given = 0;
 	vpninfo->validate_peer_cert = validate_peer_cert;
+	vpninfo->process_auth_form = process_auth_form;
 	vpninfo->cbdata = vpninfo;
 	vpninfo->cert_expire_warning = 60 * 86400;
 	vpninfo->vpnc_script = DEFAULT_VPNCSCRIPT;
@@ -469,6 +520,9 @@ int main(int argc, char **argv)
 		case OPT_PRINTCOOKIE:
 			cookieonly = 2;
 			break;
+		case OPT_AUTHENTICATE:
+			cookieonly = 3;
+			break;
 		case OPT_COOKIE_ON_STDIN:
 			read_stdin(&vpninfo->cookie);
 			/* If the cookie is empty, ignore it */
@@ -484,7 +538,6 @@ int main(int argc, char **argv)
 			break;
 		case OPT_NON_INTER:
 			non_inter = 1;
-			disable_openssl_ui();
 			break;
 		case OPT_RECONNECT_TIMEOUT:
 			vpninfo->reconnect_timeout = atoi(config_arg);
@@ -502,27 +555,14 @@ int main(int argc, char **argv)
 			vpninfo->cookie = keep_config_arg();
 			break;
 		case 'c':
-			vpninfo->cert = keep_config_arg();
+			vpninfo->cert = strdup(config_arg);
 			break;
 		case 'e':
 			vpninfo->cert_expire_warning = 86400 * atoi(config_arg);
 			break;
 		case 'k':
-			vpninfo->sslkey = keep_config_arg();
+			vpninfo->sslkey = strdup(config_arg);
 			break;
-		case 'K':
-			if (!strcasecmp(config_arg, "PKCS#12") ||
-			    !strcasecmp(config_arg, "PKCS12")) {
-				vpninfo->cert_type = CERT_TYPE_PKCS12;
-			} else if (!strcasecmp(config_arg, "TPM")) {
-				vpninfo->cert_type = CERT_TYPE_TPM;
-			} else if (!strcasecmp(config_arg, "PEM")) {
-				vpninfo->cert_type = CERT_TYPE_PEM;
-			} else {
-				fprintf(stderr, _("Unknown certificate type '%s'\n"),
-					config_arg);
-				usage();
-			}
 		case 'd':
 			vpninfo->deflate = 1;
 			break;
@@ -542,14 +582,21 @@ int main(int argc, char **argv)
 			use_syslog = 1;
 			break;
 		case 'm':
-			vpninfo->mtu = atol(config_arg);
-			if (vpninfo->mtu < 576) {
-				fprintf(stderr, _("MTU %d too small\n"), vpninfo->mtu);
-				vpninfo->mtu = 576;
+			vpninfo->reqmtu = atol(config_arg);
+			if (vpninfo->reqmtu < 576) {
+				fprintf(stderr, _("MTU %d too small\n"), vpninfo->reqmtu);
+				vpninfo->reqmtu = 576;
+			}
+			break;
+		case OPT_BASEMTU:
+			vpninfo->basemtu = atol(config_arg);
+			if (vpninfo->basemtu < 576) {
+				fprintf(stderr, _("MTU %d too small\n"), vpninfo->basemtu);
+				vpninfo->basemtu = 576;
 			}
 			break;
 		case 'p':
-			vpninfo->cert_password = keep_config_arg();
+			vpninfo->cert_password = strdup(config_arg);
 			break;
 		case 'P': 
 			proxy = keep_config_arg();
@@ -558,6 +605,7 @@ int main(int argc, char **argv)
 		case OPT_NO_PROXY:
 			autoproxy = 0;
 			proxy = NULL;
+			break;
 		case OPT_LIBPROXY:
 			autoproxy = 1;
 			proxy = NULL;
@@ -630,6 +678,7 @@ int main(int argc, char **argv)
 			break;
 		case 'V':
 			printf(_("OpenConnect version %s\n"), openconnect_version_str);
+			print_build_opts();
 			exit(0);
 		case 'x':
 			vpninfo->xmlconfig = keep_config_arg();
@@ -645,6 +694,9 @@ int main(int argc, char **argv)
 		case OPT_FORCE_DPD:
 			vpninfo->dtls_times.dpd = vpninfo->ssl_times.dpd = atoi(config_arg);
 			break;
+		case OPT_DTLS_LOCAL_PORT:
+			vpninfo->dtls_local_port = atoi(config_arg);
+			break;
 		default:
 			usage();
 		}
@@ -717,15 +769,32 @@ int main(int argc, char **argv)
 #endif
 
 	if (!vpninfo->cookie && openconnect_obtain_cookie(vpninfo)) {
+		if (vpninfo->csd_scriptname) {
+			unlink(vpninfo->csd_scriptname);
+			vpninfo->csd_scriptname = NULL;
+		}
 		fprintf(stderr, _("Failed to obtain WebVPN cookie\n"));
 		exit(1);
 	}
 
-	if (cookieonly) {
+	if (cookieonly == 3) {
+		/* --authenticate */
+		printf("COOKIE='%s'\n", vpninfo->cookie);
+		printf("HOST='%s'\n", vpninfo->hostname);
+		if (vpninfo->peer_cert) {
+			char buf[41] = {0, };
+			openconnect_get_cert_sha1(vpninfo, vpninfo->peer_cert, buf);
+			printf("FINGERPRINT='%s'\n", buf);
+		}
+		openconnect_vpninfo_free(vpninfo);
+		exit(0);
+	} else if (cookieonly) {
 		printf("%s\n", vpninfo->cookie);
-		if (cookieonly == 1)
+		if (cookieonly == 1) {
 			/* We use cookieonly=2 for 'print it and continue' */
+			openconnect_vpninfo_free(vpninfo);
 			exit(0);
+		}
 	}
 	if (make_cstp_connection(vpninfo)) {
 		fprintf(stderr, _("Creating SSL connection failed\n"));
@@ -819,7 +888,7 @@ static int write_new_config(void *_vpninfo, char *buf, int buflen)
 
 		return -err;
 	}
-	  
+
 	return 0;
 }
 
@@ -828,6 +897,9 @@ void write_progress(void *_vpninfo, int level, const char *fmt, ...)
 	FILE *outf = level ? stdout : stderr;
 	va_list args;
 
+	if (cookieonly)
+		outf = stderr;
+
 	if (verbose >= level) {
 		va_start(args, fmt);
 		vfprintf(outf, fmt, args);
@@ -874,15 +946,15 @@ void syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
 
 struct accepted_cert {
 	struct accepted_cert *next;
-	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
+	char fingerprint[SHA1_SIZE * 2 + 1];
 	char host[0];
 } *accepted_certs;
 
-static int validate_peer_cert(void *_vpninfo, X509 *peer_cert,
+static int validate_peer_cert(void *_vpninfo, OPENCONNECT_X509 *peer_cert,
 			      const char *reason)
 {
 	struct openconnect_info *vpninfo = _vpninfo;
-	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
+	char fingerprint[SHA1_SIZE * 2 + 1];
 	struct accepted_cert *this;
 	int ret;
 
@@ -898,33 +970,25 @@ static int validate_peer_cert(void *_vpninfo, X509 *peer_cert,
 		    !strcasecmp(this->fingerprint, fingerprint))
 			return 0;
 	}
-	
-	while (1) {
-		UI *ui;
-		char buf[6];
-		int ret;
 
-		fprintf(stderr,
-			_("\nCertificate from VPN server \"%s\" failed verification.\n"
-			  "Reason: %s\n"), vpninfo->hostname, reason);
-		fflush(stderr);
+	while (1) {
+		char buf[80];
+		char *details;
+		char *p;
+
+		fprintf(stderr, _("\nCertificate from VPN server \"%s\" failed verification.\n"
+			 "Reason: %s\n"), vpninfo->hostname, reason);
 
 		if (non_inter)
 			return -EINVAL;
 
-		ui = UI_new();
-		if (!ui) {
-			fprintf(stderr, _("Failed to create UI\n"));
+		fprintf(stderr, _("Enter '%s' to accept, '%s' to abort; anything else to view: "),
+		       _("yes"), _("no"));
+		if (!fgets(buf, sizeof(buf), stdin))
 			return -EINVAL;
-		}
-		UI_add_input_string(ui, _("Enter 'yes' to accept, 'no' to abort; anything else to view: "),
-				    UI_INPUT_FLAG_ECHO, buf, 2, 5);
-		ret = UI_process(ui);
-		UI_free(ui);
-		if (ret == -2)
-			return -EINVAL;
-		if (ret == -1)
-			buf[0] = 0;
+		p = strchr(buf, '\n');
+		if (p)
+			*p = 0;
 
 		if (!strcasecmp(buf, _("yes"))) {
 			struct accepted_cert *newcert = malloc(sizeof(*newcert) +
@@ -940,8 +1004,193 @@ static int validate_peer_cert(void *_vpninfo, X509 *peer_cert,
 		if (!strcasecmp(buf, _("no")))
 			return -EINVAL;
 
-		X509_print_fp(stderr, peer_cert);
+		details = openconnect_get_cert_details(vpninfo, peer_cert);
+		fputs(details, stderr);
+		free(details);
 		fprintf(stderr, _("SHA1 fingerprint: %s\n"), fingerprint);
 	}
-				
+}
+
+
+/* Return value:
+ *  < 0, on error
+ *  = 0, when form was parsed and POST required
+ *  = 1, when response was cancelled by user
+ */
+static int process_auth_form(void *_vpninfo,
+			     struct oc_auth_form *form)
+{
+	struct openconnect_info *vpninfo = _vpninfo;
+	struct oc_form_opt *opt;
+	char response[1024];
+	char *p;
+
+	if (form->banner)
+		fprintf(stderr, "%s\n", form->banner);
+
+	if (form->error)
+		fprintf(stderr, "%s\n", form->error);
+
+	if (form->message)
+		fprintf(stderr, "%s\n", form->message);
+
+	/* scan for select options first so they are displayed first */
+	for (opt = form->opts; opt; opt = opt->next) {
+		if (opt->type == OC_FORM_OPT_SELECT) {
+			struct oc_form_opt_select *select_opt = (void *)opt;
+			struct oc_choice *choice = NULL;
+			int i;
+
+			if (!select_opt->nr_choices)
+				continue;
+
+			if (vpninfo->authgroup &&
+			    !strcmp(opt->name, "group_list")) {
+				for (i = 0; i < select_opt->nr_choices; i++) {
+					choice = &select_opt->choices[i];
+
+					if (!strcmp(vpninfo->authgroup,
+						    choice->label)) {
+						opt->value = choice->name;
+						break;
+					}
+				}
+				if (!opt->value)
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Auth choice \"%s\" not available\n"),
+						     vpninfo->authgroup);
+			}
+			if (!opt->value && select_opt->nr_choices == 1) {
+				choice = &select_opt->choices[0];
+				opt->value = choice->name;
+			}
+			if (opt->value) {
+				select_opt = NULL;
+				continue;
+			}
+			if (non_inter) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("User input required in non-interactive mode\n"));
+				goto err;
+			}
+			fprintf(stderr, "%s [", opt->label);
+			for (i = 0; i < select_opt->nr_choices; i++) {
+				choice = &select_opt->choices[i];
+				if (i)
+					fprintf(stderr, "|");
+
+				fprintf(stderr, "%s", choice->label);
+			}
+			fprintf(stderr, "]:");
+			fflush(stderr);
+
+			if (!fgets(response, sizeof(response), stdin) || !strlen(response))
+				goto err;
+
+			p = strchr(response, '\n');
+			if (p)
+				*p = 0;
+
+			for (i = 0; i < select_opt->nr_choices; i++) {
+				choice = &select_opt->choices[i];
+
+				if (!strcmp(response, choice->label)) {
+					select_opt->form.value = choice->name;
+					break;
+				}
+			}
+			if (!select_opt->form.value) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Auth choice \"%s\" not valid\n"),
+					     response);
+				goto err;
+			}
+		}
+	}
+
+	for (opt = form->opts; opt; opt = opt->next) {
+
+		if (opt->type == OC_FORM_OPT_TEXT) {
+			if (vpninfo->username &&
+			    !strcmp(opt->name, "username")) {
+				opt->value = strdup(vpninfo->username);
+				if (!opt->value)
+					goto err;
+			} else if (non_inter) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("User input required in non-interactive mode\n"));
+				goto err;
+			} else {
+				opt->value=malloc(80);
+				if (!opt->value)
+					goto err;
+
+				fprintf(stderr, "%s", opt->label);
+				fflush(stderr);
+
+				if (!fgets(opt->value, 80, stdin) || !strlen(opt->value))
+					goto err;
+
+				p = strchr(opt->value, '\n');
+				if (p)
+					*p = 0;
+			}
+
+		} else if (opt->type == OC_FORM_OPT_PASSWORD) {
+			if (vpninfo->password &&
+			    !strcmp(opt->name, "password")) {
+				opt->value = vpninfo->password;
+				vpninfo->password = NULL;
+				if (!opt->value)
+					goto err;
+			} else if (non_inter) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("User input required in non-interactive mode\n"));
+				goto err;
+			} else {
+				struct termios t;
+				opt->value=malloc(80);
+				if (!opt->value)
+					goto err;
+
+				fprintf(stderr, "%s", opt->label);
+				fflush(stderr);
+
+				tcgetattr(0, &t);
+				t.c_lflag &= ~ECHO;
+				tcsetattr(0, TCSANOW, &t);
+
+				p = fgets(opt->value, 80, stdin);
+
+				t.c_lflag |= ECHO;
+				tcsetattr(0, TCSANOW, &t);
+				fprintf(stderr, "\n");
+
+				if (!p || !strlen(opt->value))
+					goto err;
+
+				p = strchr(opt->value, '\n');
+				if (p)
+					*p = 0;
+			}
+
+		}
+	}
+
+	if (vpninfo->password) {
+		free(vpninfo->password);
+		vpninfo->password = NULL;
+	}
+
+	return 0;
+
+ err:
+	for (opt = form->opts; opt; opt = opt->next) {
+		if (opt->value && (opt->type == OC_FORM_OPT_TEXT ||
+				   opt->type == OC_FORM_OPT_PASSWORD)) {
+			free(opt->value);
+			opt->value = NULL;
+		}
+	}
+	return -EINVAL;
 }
