@@ -26,12 +26,11 @@
 #include <poll.h>
 #include <limits.h>
 #include <sys/select.h>
+#include <stdlib.h>
 #include <signal.h>
 #include <unistd.h>
 #include <string.h>
 
-#include <openssl/ssl.h>
-
 #include "openconnect-internal.h"
 
 void queue_packet(struct pkt **q, struct pkt *new)
@@ -79,7 +78,7 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 		struct timeval tv;
 		fd_set rfds, wfds, efds;
 
-#ifdef SSL_OP_CISCO_ANYCONNECT
+#ifdef HAVE_DTLS
 		if (vpninfo->new_dtls_ssl)
 			dtls_try_handshake(vpninfo);
 
