@@ -30,10 +30,20 @@
 #include <sys/types.h>
 #include <unistd.h>
 
-#define OPENCONNECT_API_VERSION_MAJOR 1
-#define OPENCONNECT_API_VERSION_MINOR 4
+#define OPENCONNECT_API_VERSION_MAJOR 2
+#define OPENCONNECT_API_VERSION_MINOR 0
 
 /*
+ * API version 2.0:
+ *  - OPENCONNECT_X509 is now an opaque type.
+ *  - Add openconnect_has_pkcs11_support(), openconnect_has_tss_blob_support()
+ *  - Rename openconnect_init_openssl() -> openconnect_init_ssl()
+ *  - Rename openconnect_vpninfo_new_with_cbdata() -> openconnect_vpninfo_new()
+ *    and kill the old openconnect_vpninfo_new() and its callback types.
+ *
+ * API version 1.5:
+ *  - Add openconnect_get_cert_details(), openconnect_get_cert_DER().
+ *
  * API version 1.4:
  *  - Add openconnect_set_cancel_fd()
  *
@@ -119,19 +129,28 @@ struct oc_auth_form {
 #define PRG_TRACE	3
 
 struct openconnect_info;
-/* We don't want to have to pull in OpenSSL stuff just for this */
-struct x509_st;
+
+#define OPENCONNECT_X509 void
+
+/* Unless otherwise specified, all functions which set strings will take
+   ownership of those strings and the library will free them later in
+   openconnect_vpninfo_free() */
 
 
-
-/* Unless otherwise specified, all functions which set strings will take ownership of those strings
-   and should free them later in openconnect_vpninfo_free() */
+/* The buffer 'buf' must be at least 41 bytes. It will receive a hex string
+   with trailing NUL, representing the SHA1 fingerprint of the certificate. */
 int openconnect_get_cert_sha1(struct openconnect_info *vpninfo,
-			      struct x509_st *cert, char *buf);
+			      OPENCONNECT_X509 *cert, char *buf);
+char *openconnect_get_cert_details(struct openconnect_info *vpninfo,
+				   OPENCONNECT_X509 *cert);
+/* Returns the length of the created DER output, in a newly-allocated buffer
+   that will need to be freed by the caller. */
+int openconnect_get_cert_DER(struct openconnect_info *vpninfo,
+			     OPENCONNECT_X509 *cert, unsigned char **buf);
 int openconnect_set_http_proxy(struct openconnect_info *vpninfo, char *proxy);
 int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo);
 int openconnect_obtain_cookie(struct openconnect_info *vpninfo);
-void openconnect_init_openssl(void);
+void openconnect_init_ssl(void);
 
 char *openconnect_get_vpn_name (struct openconnect_info *);
 char *openconnect_get_hostname (struct openconnect_info *);
@@ -140,8 +159,10 @@ char *openconnect_get_urlpath (struct openconnect_info *);
 void openconnect_set_urlpath (struct openconnect_info *, char *);
 
 /* This function does *not* take ownership of the string; it's copied
-   into a static buffer in the vpninfo */
-void openconnect_set_xmlsha1 (struct openconnect_info *, char *, int size);
+   into a static buffer in the vpninfo. The size must be 41 bytes,
+   since that's the size of a 20-byte SHA1 represented as hex with
+   a trailing NUL. */
+void openconnect_set_xmlsha1 (struct openconnect_info *, const char *, int size);
 
 void openconnect_set_cafile (struct openconnect_info *, char *);
 void openconnect_setup_csd (struct openconnect_info *, uid_t, int silent, char *wrapper);
@@ -151,7 +172,7 @@ void openconnect_set_client_cert (struct openconnect_info *, char *cert, char *s
  * will be valid when a cookie has been obtained successfully, and will
  * be valid until the connection is destroyed or another attempt it made
  * to use it. */
-struct x509_st *openconnect_get_peer_cert (struct openconnect_info *);
+OPENCONNECT_X509 *openconnect_get_peer_cert (struct openconnect_info *);
 
 int openconnect_get_port (struct openconnect_info *);
 char *openconnect_get_cookie (struct openconnect_info *);
@@ -188,8 +209,8 @@ const char *openconnect_get_version(void);
    if the certificate is (or has in the past been) explicitly accepted
    by the user, and non-zero to abort the connection. */
 typedef int (*openconnect_validate_peer_cert_vfn) (void *privdata,
-						  struct x509_st *cert,
-						  const char *reason);
+						   OPENCONNECT_X509 *cert,
+						   const char *reason);
 /* On a successful connection, the server may provide us with a new XML
    configuration file. This contains the list of servers that can be
    chosen by the user to connect to, amongst other stuff that we mostly
@@ -199,14 +220,19 @@ typedef int (*openconnect_validate_peer_cert_vfn) (void *privdata,
    then the new XML is downloaded and this function is invoked. */
 typedef int (*openconnect_write_new_config_vfn) (void *privdata, char *buf,
 						int buflen);
-/* Handle an authentication form, requesting input from the user. */
+/* Handle an authentication form, requesting input from the user. 
+ * Return value:
+ *  < 0, on error
+ *  = 0, when form was parsed and POST required
+ *  = 1, when response was cancelled by user
+ */
 typedef int (*openconnect_process_auth_form_vfn) (void *privdata,
 						 struct oc_auth_form *form);
 /* Logging output which the user *may* want to see. */
 typedef void __attribute__ ((format(printf, 3, 4)))
 		(*openconnect_progress_vfn) (void *privdata, int level,
 					    const char *fmt, ...);
-struct openconnect_info *openconnect_vpninfo_new_with_cbdata (char *useragent,
+struct openconnect_info *openconnect_vpninfo_new (char *useragent,
 						  openconnect_validate_peer_cert_vfn,
 						  openconnect_write_new_config_vfn,
 						  openconnect_process_auth_form_vfn,
@@ -214,24 +240,13 @@ struct openconnect_info *openconnect_vpninfo_new_with_cbdata (char *useragent,
 						  void *privdata);
 void openconnect_vpninfo_free (struct openconnect_info *vpninfo);
 
-/* Old prototypes of the callback functions, which used to be passed the
-   vpninfo instead of a caller-provided pointer. You probably don't want to
-   use these; they're here for compatibility only. */
-typedef int (*openconnect_validate_peer_cert_fn) (struct openconnect_info *,
-						  struct x509_st *cert,
-						  const char *reason);
-typedef int (*openconnect_write_new_config_fn) (struct openconnect_info *, char *buf,
-						int buflen);
-typedef int (*openconnect_process_auth_form_fn) (struct openconnect_info *,
-						 struct oc_auth_form *form);
-typedef void __attribute__ ((format(printf, 3, 4)))
-		(*openconnect_progress_fn) (struct openconnect_info *, int level,
-					    const char *fmt, ...);
-/* Don't use this. Use the _with_cbdata version instead. */
-struct openconnect_info *openconnect_vpninfo_new (char *useragent,
-						  openconnect_validate_peer_cert_fn,
-						  openconnect_write_new_config_fn,
-						  openconnect_process_auth_form_fn,
-						  openconnect_progress_fn);
+/* SSL certificate capabilities. openconnect_has_pkcs11_support() means that we
+   can accept PKCS#11 URLs in place of filenames, for the certificate and key. */
+int openconnect_has_pkcs11_support(void);
+
+/* The OpenSSL TPM ENGINE stores keys in a PEM file labelled with the string
+   -----BEGIN TSS KEY BLOB-----. GnuTLS may learn to support this format too,
+   in the near future. */
+int openconnect_has_tss_blob_support(void);
 
 #endif /* __OPENCONNECT_H__ */
